"""
Module adapted and copied from:
https://github.com/Ch00k/ffmpy

"""

import errno
import logging
import subprocess


class FFmpeg(object):
    """Wrapper for `FFmpeg <https://www.ffmpeg.org/>`.

    Args:
        source     (str): path to source file
        target     (str): path to target file
        options    (str): string containing options for ffmpeg as use in
            console (arguments separated by space)
        executable (str, optional): path to ffmpeg executable. Defaults to
            'ffmpeg': Can be overwritten in case e.g libav is used.
    """

    def __init__(self, source, target, options=None, executable="ffmpeg"):
        """Initialize ffmpeg command line wrapper.

        """
        self.executable = executable
        self._cmd = [executable]

        self._cmd.extend(["-i", source])
        if options:
            self._cmd.extend(options.split())
        self._cmd.extend([target])

        self.cmd = subprocess.list2cmdline(self._cmd)
        self.process = None

    def __repr__(self):
        return "<{0!r} {1!r}>".format(self.__class__.__name__, self.cmd)

    def run(self):
        """Execute ffmpeg command line. Log stderr output.

        Raises:
            FFRuntimeError           : in case ffmpeg command exits with a
                non-zero code.
            FFExecutableNotFoundError: in case the executable path passed
                as not valid.
        """

        try:
            process_result = subprocess.run(self._cmd, stderr=subprocess.PIPE)
        except OSError as e:
            if e.errno == errno.ENOENT:
                logging.error(
                    "ffmpeg path not found. either ffmpeg is not "
                    "installed are not at the standard path."
                )
                raise FFExecutableNotFoundError(
                    "Executable '{0}' not found".format(self.executable)
                )
            else:
                raise

        catchers = ["Input #0", "Stream #0:", "Output #0"]

        for line in process_result.stderr.decode().split("\n"):
            for catch in catchers:
                if catch in line:
                    logging.log(25, line.strip())

        logging.debug(process_result.stderr.decode())

        if process_result.returncode != 0:
            logging.error(
                "command \n{}\n produced the following error:\n {}".format(
                    self.cmd, process_result.stderr
                )
            )
            raise FFRuntimeError(
                self.cmd,
                process_result.returncode,
                process_result.stdout,
                process_result.stderr,
            )


class FFExecutableNotFoundError(Exception):
    """Raise when ffmpeg executable was not found."""


class FFRuntimeError(Exception):
    """Raise when ffmpeg command line execution returns a non-zero exit code.

    """

    def __init__(self, cmd, exit_code, stdout, stderr):
        self.cmd = cmd
        self.exit_code = exit_code
        self.stdout = stdout
        self.stderr = stderr

        message = "`{0}` exited with status {1}\n\nSTDOUT:\n{2}\n\nSTDERR:\n{3}".format(
            self.cmd, exit_code, (stdout or b"").decode(), (stderr or b"").decode()
        )

        super(FFRuntimeError, self).__init__(message)



import logging


def init(path_config):
    """initialize all modules, which must be initialized

    can be done individually too.

    Args:
        path_config (str): path to allocation map config yaml file. Can be
            None, use at own risk!
    """
    init_logging()
    logging.log(25, "testing mmusicc log level. remove in future commit")
    init_formats()
    if path_config:
        init_allocationmap(path_config)


def init_allocationmap(path_config):
    """initialize allocation map of tags.

    Args:
        path_config (str): path to allocation map config yaml file
    """
    from mmusicc.util import allocationmap

    allocationmap.init_allocationmap(path_config)


def init_formats():
    """initialize formats module"""
    from mmusicc.formats import init

    init()


def init_logging(level=25):
    """initialize logger"""
    from mmusicc.util import init_logger

    init_logger(level=level)



SPLIT_CHAR = ["/", "\n", ";"]
JOIN_CHAR = "; "


def join_str_list(str_list):
    """joins a list of strings with a globally defined separator

    used for recombining multiple values in one tag field (e.g. Composer).

    Args:
        str_list (list<str>): list of strings to be joined
    """
    return JOIN_CHAR.join([s.strip() for s in str_list])


def text_parser_get(text):
    """splits a text string at globally defined split chars (recursive with lists).

    Also strips whitespaces from the strings.

    Args:
        text (str or list<str>): source string to be split. Or source list
            which contents to be split.
    Returns:
        str OR list<str>: String when single value in string else list of
            values in string.
    """
    if isinstance(text, list):
        tmp_text = list()
        for t in text:
            tmp_text.append(text_parser_get(t))
        return tmp_text
    elif isinstance(text, str):
        for c in SPLIT_CHAR:
            tmp_text = text.split(c)
            if len(tmp_text) > 1:
                return [t.strip() for t in tmp_text]
        else:
            return text.strip()
    else:
        raise ValueError("text wrong value")



import logging
import mimetypes
import pathlib

import mmusicc.util.allocationmap as am
from mmusicc.formats import mimes


def is_supported_audio(file):
    """Return True if file is a supported audio file."""
    mimetype = mimetypes.guess_type(str(file))

    if mimetype[0] in mimes:
        return True
    else:
        return False


def process_white_and_blacklist(whitelist, blacklist):
    """creates a whitelist from one whitelist and one blacklist.

    Blacklist is processed after whitelist and will remove whitelisted
    items.

    Args:
        whitelist (list of str): whitelist of tags to be imported. Loads all
            tags if None.
        blacklist (list of str): blacklist of tags not to be imported. These
            items are removed from the whitelist. If none, whitelist is
            returned unprocessed.

    Returns:
        list<str>: whitelist after applying blacklisting.
    """
    if not whitelist:
        whitelist = am.list_tags.copy()
    if blacklist:
        for t in blacklist:
            try:
                whitelist.pop(whitelist.index(t))
            except ValueError:
                logging.warning(
                    "Warning can not remove {}. "
                    "It is not in the whitelist.".format(t)
                )
                continue
    return whitelist


def swap_base(root_a, path_a, root_b):
    """removes the root part of path_a and attaches the remainder to root
    of root_b crating path_b.

    Args:
        root_a (pathlib.Path): Root path of a.
        path_a (pathlib.Path): Full path of a.
        root_b (pathlib.Path): Root path of b.

    Returns:
        (pathlib.Path): Full path of path_b.

    """
    relative_path = path_a.relative_to(root_a)
    path_b = root_b.joinpath(relative_path)
    return path_b


def get_the_right_one(list_path, match_path, drop_suffix=True):
    """return the most matching path out of a list of (absolute) paths

    with has the most common elements beginning at the leaves of the tree.
    Basically finds matching relative paths with unknown working directory.
    All string paths are automatically converted to pathlib Paths.

    Args:
        list_path (list of pathlib.Path or list of str):
            list of paths to be compared.
        match_path                (str or pathlib.Path):
            path to be matched.
        drop_suffix                    (bool, optional):
            if True, compare paths without extensions (if path points to a
            file). Defaults to True.

    Returns:
        pathlib.Path: Path from list which matches the match_path the most.

    Exceptions:
        KeyError: if not matching entry has been found.

    Example:
        >>> list_path = ['fuu/bar/mmusicc.flac', 'fuu/rab/mmusicc.ogg']
        >>> match_path = 'root/bar/mmusicc.ogg'
        >>> get_the_right_one(list_path, match_path)
        'fuu/bar/mmusicc.flac'
    """

    list_path = list_path.copy()
    for i in range(len(list_path)):
        list_path[i] = pathlib.Path(list_path[i])

    if not isinstance(match_path, pathlib.Path):
        match_path = pathlib.Path(match_path)
    if drop_suffix:
        match_path = match_path.with_suffix("")

    match_parts = match_path.parts

    i = 0
    while len(list_path) > 1:
        i += 1
        list_new = list()
        for path in list_path:
            if path.with_suffix("").parts[-i] == match_parts[-i]:
                list_new.append(path)
            else:
                pass
        list_path = list_new

        if i > 100:
            raise Exception("loop count to high")

    if len(list_path) == 0:
        raise KeyError("could not find matching entry")

    return list_path[0]



#  Copyright (c) 2020 Johannes Nolte
#  SPDX-License-Identifier: GPL-3.0-or-later

import logging
import os
import pathlib
import re

import mmusicc.util.allocationmap as am
from mmusicc.database import MetaDB
from mmusicc.formats import MusicFile
from mmusicc.util.metadatadict import MetadataDict, Empty, Div
from mmusicc.util.misc import (
    is_supported_audio,
    process_white_and_blacklist,
    get_the_right_one,
)


class MetadataMeta(type):
    """Meta Object for Metadata class."""

    def __init__(cls, name, bases, nmspc):
        super(MetadataMeta, cls).__init__(name, bases, nmspc)
        cls._dry_run = False
        cls._database = None
        # cls._delete_existing = False

    @property
    def dry_run(cls):
        """bool: Get or set. If True do everything as usual, but without
        writing data.
        """
        if cls._dry_run:
            logging.log(25, "Running in Dry Run mode. " "No data will be written.")
        return cls._dry_run

    @dry_run.setter
    def dry_run(cls, value):
        cls._dry_run = value

    @property
    def is_linked_database(cls):
        """bool: Get True if a database is linked to class."""
        if cls._database:
            return True
        return False

    def link_database(cls, database_url):
        """Link a database to class.

        Args:
            database_url (str): database url following RFC-1738*. If the sting,
                does not contain '://', a filepath for a sqlite database is
                assumed.

        Raises:
            Exception: if a database is already linked
        """
        if not cls._database:
            cls._database = MetaDB(database_url)
        else:
            raise Exception("Database already linked")

    def unlink_database(cls):
        """Unlink the database from the class.

        Raises:
            Exception: if no database is linked
        """
        if cls._database:
            cls._database = None
        else:
            raise Exception("No Database linked")

    # prep for future class property
    # @property
    # def delete_existing(cls):
    #     return cls._delete_existing
    #
    # @delete_existing.setter
    # def delete_existing(cls, value):
    #     cls._delete_existing = value
    #     if cls._dry_run:
    #         logging.log(25, "Existing Metadata will be deleted!")


class Metadata(metaclass=MetadataMeta):
    """Class containing Metadata Information,

    either from a linked file or loaded from a database.

    Args:
        file_path (str or pathlib.Path, optional): path to an supported audio
            file, can be set later with 'link_audio_file()' too. Defaults to
            None.
        read_tag (bool, optional): enables automatic reading of metadata from
            file at class initialisation. Defaults to True.
    """

    def __init__(self, file_path=None, read_tag=True):

        if not am.list_tags:
            raise Exception("mmusicc not initialized! Please Initialize first")

        self._audio = None

        self._dict_data = MetadataDict()

        if file_path:
            self.link_audio_file(file_path)

            if read_tag:
                self.read_tags()

        self.dict_auto_fill_org = None

    @property
    def file_path(self):
        """pathlib.Path: Get file path of linked audio file."""
        if self._audio:
            return self._audio.file_path
        return None

    @property
    def audio_file_linked(self):
        """bool: Get True if a audio file is linked to instance."""
        if self._audio:
            return True
        return False

    def link_audio_file(self, file_path):
        """Links audio file with given path to instance.

        Args:
            file_path (str or pathlib.Path): file path of audio file.

        Raises:
            FileNotFoundError: if file does not exist.
        """
        file_path = pathlib.Path(file_path)
        if file_path.exists():
            self._audio = MusicFile(file_path)
        else:
            raise FileNotFoundError("Error Audio File does not exist")

    @property
    def unprocessed_tag(self):
        """pathlib.Path: Get file path of linked audio file."""
        if self._audio:
            return self._audio.unprocessed_tag
        return None

    @property
    def dict_data(self):
        """MetadataDict: Get the dict containing tags"""
        return self._dict_data

    def get_tag(self, str_tag_key):
        """Get value of given str_tag_key"""
        return self._dict_data.get(str_tag_key)

    def set_tag(self, str_tag_key, value):
        """Set value of tag with given str_tag_key"""
        self._dict_data[str_tag_key] = value

    def read_tags(self):
        """Read metadata from linked audio file into dict meta.

        Raises:
            Exception: if no file is linked
        """
        if not self._audio:
            raise Exception("no file_path linked")
        self._audio.file_read()
        self._dict_data.update(self._audio.dict_meta)

    def write_tags(self, remove_existing=False, write_empty=False):
        """write metadata to linked audio file

        Args:
            remove_existing (bool, optional): If true clear all tags on file
                before writing. Defaults to False.
            write_empty (bool): if true write empty tags, exact effect
                depends on comment type. Either the tag entries will not exist
                or overwritten with None/Null/"". Defaults to False.
        Raises:
            Exception: if no file is linked
        """
        if not self._audio:
            raise Exception("no file_path linked")
        self._audio.dict_meta.update(self._dict_data)
        if not Metadata.dry_run:
            self._audio.file_save(
                remove_existing=remove_existing, write_empty=write_empty
            )

    def import_tags(self, source_meta, whitelist=None, blacklist=None, skip_none=True):
        """Imports metadata from another Metadata object.

        Args:
            source_meta (Metadata): Metadata object containing the
                tags to be imported.
            whitelist (list of str, optional): whitelist of tags to be
                imported. If None, loads all tags (except blacklisted).
                Defaults to None.
            blacklist (list of str, optional): blacklist of tags not to be
                imported. Applied after whitelist. If None, no tags are
                blacklisted. Defaults to None.
            skip_none (bool, optional): If True, don't overwrite values in
                target, which are None in source. Defaults to True.
        """
        self._import_tags(source_meta._dict_data, whitelist, blacklist, skip_none)

    def _import_tags(self, dict_meta, whitelist, blacklist, skip_none):
        tags = process_white_and_blacklist(whitelist, blacklist)
        for tag in am.list_tags:
            if tag in tags:
                val = dict_meta[tag]
                if not val and skip_none:
                    continue
                self._dict_data[tag] = val

    def import_tags_from_db(
        self, primary_key=None, whitelist=None, blacklist=None, skip_none=True
    ):
        """Imports metadata from the database.

        Args:
            primary_key (str, None): unique identifier of the item
                which data has to be loaded. The save function only uses the
                absolute filepath atm. If value is None, a algorithm takes the
                path of the linked file works and works itself backward
                (beginning at the leave) in the key list of the DB until only
                one key is left, which is used. In other words, its acts like
                the keys are relative file path (with unknown working
                directory). Defaults to None.
            whitelist (list of str, optional): whitelist of tags to be
                imported. If None, loads all tags (except blacklisted).
                Defaults to None.
            blacklist (list of str, optional): blacklist of tags not to be
                imported. Applied after whitelist. If None, no tags are
                blacklisted. Defaults to None.
            skip_none (bool, optional): If True, don't overwrite values in
                target, which are None in source. Defaults to True.

            Raises:
             Exception: if no database linked to class

            """

        if not primary_key and self.file_path:
            keys = self._database.get_list_of_primary_keys()
            primary_key = get_the_right_one(keys, self.file_path)

        if not self._database:
            raise Exception("no database linked")
        else:
            data = self._database.read_meta(str(primary_key))
            if data:
                self._import_tags(data, whitelist, blacklist, skip_none)
            else:
                logging.warning(
                    "database read failed, no data imported. "
                    "File might not be in database"
                )

    def export_tags_to_db(self, root_dir=None):
        """Saves all tags to database.

         This is the secure way. Data not wanted does not have to be loaded,
         but all data can still be accessed in case it is needed again.

         Raises:
             Exception: if no database linked to class
         """
        if not self._database:
            raise Exception("no database linked")
        if self.file_path:
            if not self._dry_run:
                primary_key = str(self.file_path)
                self._database.insert_meta(self._dict_data, primary_key)
        else:
            pass

    def auto_fill_tags(self):
        """Automatic fill/autocomplete tags with in config file defined rules.

        This future is intended to fix small consistency errors in metadata,
        like missing album artists or leading 0 before a single digit
        tracknumber.
        """
        if not self.dict_auto_fill_org:
            self.dict_auto_fill_org = MetadataDict(init_value=False)
        for tag in list(am.dict_auto_fill_rules):
            rule = am.dict_auto_fill_rules.get(tag)
            val_test = self._dict_data.get(rule[0])
            val_regex = rule[1]
            try:
                val_parse = self._dict_data.get(rule[2])
            except KeyError:
                val_parse = rule[2]
            if val_regex is None:
                if isinstance(val_test, Empty):
                    self.dict_auto_fill_org[tag] = self._dict_data[tag]
                    self._dict_data[tag] = val_parse
            elif isinstance(val_regex, str):
                if not isinstance(val_test, str):
                    val_test = str(val_test)
                m = re.search(val_regex, val_test)
                if m:
                    self.dict_auto_fill_org[tag] = self._dict_data[tag]
                    self._dict_data[tag] = eval(val_parse)


class GroupMetadata(Metadata):
    """Class holding one ore many Metadata Objects. Subclasses Metadata and

    overwrites Metadata functions so that you don't have to care if you have
    one file or a list of files like an album. Tags in dict are a summary of
    the tags of individual values of the contained metadata. If all values of a
    tag are identical the value is in dict as it is, otherwise a Div object is
    created managing different values.

    Args:
        list_metadata (list of Metadata or list of str or list of pathlib.Path):
            list of Metadata objects or file paths.

    Note:
        Not all properties and functions of Metadata are supported (like
        file_path, obviously). Stick with the documented ones.
    """

    def __init__(self, list_metadata):
        super().__init__(None)
        if isinstance(list_metadata[0], Metadata):
            self.list_metadata = list_metadata
        else:
            self.list_metadata = list()
            for file_path in list_metadata:
                if is_supported_audio(file_path):
                    self.list_metadata.append(Metadata(file_path))
                else:
                    logging.warning("File '{}' is no audio file".format(file_path))
        self.read_tags()

        self.dict_auto_fill_org = None

    def get_tag(self, str_tag_key):
        """Super-Method applied to all Objects in list. See Metadata."""
        return self._dict_data.get(str_tag_key)

    def set_tag(self, str_tag_key, value):
        """Super-Method applied to all Objects in list. See Metadata."""
        for metadata in self.list_metadata:
            metadata._dict_data[str_tag_key] = value
        self.__compare_tags()

    def reset_meta(self):
        for metadata in self.list_metadata:
            metadata._dict_data.reset()

    def auto_fill_tags(self):
        """Super-Method applied to all Objects in list. See Metadata."""
        if not self.dict_auto_fill_org:
            self.dict_auto_fill_org = MetadataDict(init_value=False)
        for metadata in self.list_metadata:
            metadata.auto_fill_tags()
        self.__compare_tags()

    def read_tags(self):
        """Super-Method applied to all Objects in list. See Metadata."""
        for metadata in self.list_metadata:
            metadata.read_tags()
        self.__compare_tags()

    def __compare_tags(self):
        """compares tag values of all objects in list and creates a Div object,
        which indicates different values (and also got a list of all the
        differences).
        """
        for key in am.list_tags:
            first = True
            for metadata in self.list_metadata:
                data = metadata._dict_data.get(key)
                if first:
                    self._dict_data[key] = data
                    first = False
                else:
                    if self._dict_data[key] == data:
                        pass
                    else:
                        self._dict_data[key] = Div(key, self.list_metadata)
                        break

    def write_tags(self, remove_existing=False, write_empty=False):
        """Super-Method applied to all Objects in list. See Metadata."""
        for metadata in self.list_metadata:
            metadata.write_tags(
                remove_existing=remove_existing, write_empty=write_empty
            )

    def import_tags(self, source_meta, whitelist=None, blacklist=None, skip_none=True):
        """Super-Method applied to all Objects in list. See Metadata.

        Args:
            source_meta     (GroupMetadata): group metadata object.
            whitelist (list<str>, optional): See Metadata.import_tags().
            blacklist (list<str>, optional): See Metadata.import_tags().
            skip_none      (bool, optional): See Metadata.import_tags().
        """
        paths = dict()
        for sm in source_meta.list_metadata:
            paths[sm.file_path] = sm
        path_keys = list(paths.keys())

        for metadata_self in self.list_metadata:
            key = get_the_right_one(path_keys, metadata_self.file_path)
            metadata_source = paths.get(key)
            metadata_self.import_tags(
                metadata_source,
                whitelist=whitelist,
                blacklist=blacklist,
                skip_none=skip_none,
            )
        self.__compare_tags()

    def import_tags_from_db(
        self, whitelist=None, blacklist=None, root_dir=None, skip_none=True
    ):
        """Super-Method applied to all Objects in list. See Metadata."""
        for metadata in self.list_metadata:
            metadata.import_tags_from_db(
                primary_key=None,
                whitelist=whitelist,
                blacklist=blacklist,
                skip_none=skip_none,
            )
        self.__compare_tags()

    def export_tags_to_db(self, root_dir=None):
        """Super-Method applied to all Objects in list. See Metadata."""
        for metadata in self.list_metadata:
            metadata.export_tags_to_db()

    @property
    def audio_file_linked(self):
        """Do not use this property. Only applies to Metadata."""
        raise NotImplementedError()

    @property
    def unprocessed_tag(self):
        """Super-Method applied to all Objects in list. See Metadata."""
        unprocessed_tags = dict()
        for metadata in self.list_metadata:
            if metadata._audio:
                unprocessed_tags.update(self._audio.unprocessed_tag)
        return unprocessed_tags

    @property
    def file_path(self):
        """Do not use this property. Only applies to Metadata."""
        raise NotImplementedError()

    def link_audio_file(self, file_path):
        """Do not use this property. Only applies to Metadata."""
        raise NotImplementedError()


class AlbumMetadata(GroupMetadata):
    """Special case of GroupMetadata where Metadata list is created from a
    Folder/Album path. Skips non audio files.

    Args:
        path_album: filepath of album or folder to be accessed as group.
    """

    def __init__(self, path_album):
        path_album = pathlib.Path(path_album)
        list_metadata = list()
        for file in os.listdir(path_album):
            file_path = path_album.joinpath(file)
            if is_supported_audio(file_path):
                list_metadata.append(file_path)
        super().__init__(list_metadata)



import sys

if __package__ is None and not hasattr(sys, "frozen"):
    # direct call of __main__.py
    import os.path

    path = os.path.realpath(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(path)))


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    import mmusicc

    mmusicc.MmusicC(args)


if __name__ == "__main__":
    main()



import argparse
import enum
import logging
import os
import pathlib
import sys
import textwrap

from mmusicc._init import init_formats, init_logging, init_allocationmap
from mmusicc.formats import loaders as audio_loader
from mmusicc.metadata import Metadata, AlbumMetadata
from mmusicc.util.allocationmap import get_tags_from_strs
from mmusicc.util.ffmpeg import FFmpeg
from mmusicc.util.misc import is_supported_audio, swap_base, process_white_and_blacklist
from mmusicc.version import __version__ as package_version

str_description_rqw = textwrap.dedent(
    """\
    metadata and file syncing the following combination are possible:
      - file   --> file
      - file   --> parent folder (target name is generated from source)
      - folder --> folder        (use --album to not to move through tree)
      - folder --> db            (full path as primary key)
      - db     --> folder        (key matching starts at leave of path)
    Supported Formats for Metadata: 
    {}
    """
)


# noinspection PyMethodMayBeStatic
class MmusicC:

    # TODO allow a predefined config file with all parser options to be passed

    def __init__(self, args):
        """Create a ViewControl object with the given options.
        Args:
            args (:list: string):
        """

        pre_parser = argparse.ArgumentParser(add_help=False)
        pre_parser.add_argument("-v", "--verbose", action="count", default=0)
        parsed, remaining = pre_parser.parse_known_args(args)
        self.pre_result_verbose = parsed.verbose

        if self.pre_result_verbose == 0:
            log_level = 25
        elif self.pre_result_verbose == 1:
            log_level = logging.INFO
        else:
            log_level = logging.DEBUG

        init_logging(log_level)
        init_formats()

        str_description = str_description_rqw.format(
            textwrap.fill(str(sorted(list(audio_loader))))
        )

        # noinspection PyTypeChecker
        self.parser = argparse.ArgumentParser(
            prog="mmusicc",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=str_description,
            # epilog="",
            add_help=False,
        )

        pg_required = self.parser.add_argument_group("Required Options")
        pg_general = self.parser.add_argument_group("General Options")
        pg_conversion = self.parser.add_argument_group("File Conversion")
        pg_meta = self.parser.add_argument_group("Metadata Syncing")

        group_source = pg_required.add_mutually_exclusive_group(required=True)
        group_source.add_argument(
            "-s", "--source", action="store", help="source file/album/lib-root"
        )
        group_source.add_argument(
            "-sdb", "--source-db", action="store", help="source database"
        )

        group_target = pg_required.add_mutually_exclusive_group(required=True)
        group_target.add_argument(
            "-t", "--target", action="store", help="target file/album/lib-root"
        )
        group_target.add_argument(
            "-tdb", "--target-db", action="store", help="target database"
        )

        pg_general.add_argument(
            "-h",
            "--help",
            action="help",
            default=argparse.SUPPRESS,
            help="Show this help message and exit.",
        )
        pg_general.add_argument("--version", action="version", version=package_version)
        pg_general.add_argument(
            "--album", action="store_true", help="only sync folder level"
        )
        group1 = pg_general.add_mutually_exclusive_group()
        group1.add_argument(
            "--only-meta",
            action="store_true",
            help="only sync meta, don't sync files. "
            "Auto set when syncing from/to database",
        )
        group1.add_argument(
            "--only-files",
            action="store_true",
            help="only sync files, don't update metadata",
        )
        pg_general.add_argument(
            "--dry-run",
            action="store_true",
            help="do everything as usual, but without writing (file and "
            "database). It is recommended to use with --only-meta or "
            "--only-files, otherwise errors are likely.",
        )
        pg_general.add_argument(
            "-v",
            "--verbose",
            action="count",
            default=0,
            help="print log messages. can be stacked up to 2 (info, debug)"
            # processed before argparser
        )
        # pg_general.add_argument(
        #     '-q', '--quiet',
        #     action='count',
        #     default='0',
        #     help='print log messages'
        # )

        pg_conversion.add_argument(
            "-f",
            "--format",
            action="store",
            required=False,
            help="output container format of ffmpeg conversion "
            "(ignored when target is file_path)",
        )
        pg_conversion.add_argument(
            "--ffmpeg-options",
            action="store",
            help="conversion options for ffmpeg conversion. "
            "If empty ffmpeg defaults are used.",
        )

        tmp_double_txt_2 = (
            "Can be passed as file (Plain text file, "
            "#containing "
            "tags separated with a new line) or as one or "
            "multiple arguments"
        )
        pg_meta.add_argument(
            "--white-list-tags",
            nargs="+",
            action="store",
            help="Tags to be whitelisted. " + tmp_double_txt_2,
        )
        pg_meta.add_argument(
            "--black-list-tags",
            nargs="+",
            action="store",
            help="Tags to be blacklisted. " + tmp_double_txt_2,
        )
        pg_meta.add_argument(
            "--lazy",
            action="store_true",
            help="don't overwrite existing value in target with None from source. "
            "Only  effects metadata import. Use '--delete-existing-metadata' to "
            "remove unwanted or empty tags.",
        )
        pg_meta.add_argument(
            "--delete-existing-metadata",
            action="store_true",
            help="delete existing metadata on target " "files before writing.",
        )
        pg_meta.add_argument(
            "--path-config", action="store", help="file path to custom config file"
        )

        logging.debug("mmusicc running with arguments: {}".format(args))

        self.result = self.parser.parse_args(args)

        Metadata.dry_run = self.result.dry_run

        if not self.result.path_config:
            path_folder_this_file = pathlib.Path(__file__).resolve().parent
            self.result.path_config = path_folder_this_file.joinpath(
                "data", "config.yaml"
            )

        try:
            init_allocationmap(self.result.path_config)
        except FileNotFoundError:
            self.parser.error("path to config file not found")

        self.run_files = not self.result.only_meta
        self.run_meta = not self.result.only_files

        self.db_url = None
        if self.result.source:
            self.source = pathlib.Path(self.result.source).expanduser().resolve()
        else:
            self.source = None
            self.db_url = self.result.source_db

        if self.result.target:
            self.target = pathlib.Path(self.result.target).expanduser().resolve()
        else:
            self.target = None
            self.db_url = self.result.target_db

        if self.db_url:
            if Metadata.is_linked_database:
                Metadata.unlink_database()
            Metadata.link_database(self.db_url)

        self.source_type = self.get_element_type(self.source)
        self.target_type = self.get_element_type(self.target)
        logging.info(
            "Source is '{}'. Target is '{}'.".format(self.source_type, self.target_type)
        )

        if not (self.source or self.target) and self.run_files:
            self.parser.error("Target or source is database! I can only run meta!")

        if not self.source and not self.target:
            self.parser.error("Can't sync from database to database!")

        if self.source and self.target:
            self.format_extension = None
            if self.target_type == MmusicC.ElementType.file:
                if "." in self.target.suffix:  # target is file
                    self.format_extension = self.target.suffix

            if not self.format_extension and self.result.format:
                # target is folder an format has to be given!
                self.format_extension = self.result.format
                if "." not in self.format_extension:
                    self.format_extension = "." + self.format_extension

            if not self.format_extension:
                self.parser.error(
                    "the following arguments are required: "
                    "-f/--format, except at file-->file and "
                    "album/lib-->database operations."
                )

        self.whitelist = None
        try:
            if self.result.white_list_tags:
                self.whitelist = load_tags_from_list_or_file(
                    self.result.white_list_tags
                )
        except KeyError as err:
            self.parser.error(
                "Whitelist: {}. If input is file it is mot found.".format(err)
            )

        self.blacklist = None
        try:
            if self.result.black_list_tags:
                self.blacklist = load_tags_from_list_or_file(
                    self.result.black_list_tags
                )
        except KeyError as err:
            self.parser.error(
                "Blacklist: {}. If input is file it is mot found.".format(err)
            )

        # just fot the log
        if self.whitelist or self.blacklist:
            whitelist = process_white_and_blacklist(self.whitelist, self.blacklist)
            logging.info("Tags to be Synced: {}".format(whitelist))

        if self.db_url:
            if self.source_type == MmusicC.ElementType.file:
                self.handle_media2db(self.source)
            elif self.source_type == MmusicC.ElementType.folder:
                if self.result.album:
                    self.handle_media2db(self.source)
                else:
                    gen = os.walk(self.source, topdown=True)
                    for root, dirs, files in gen:
                        audio_files = [
                            file for file in files if is_supported_audio(file)
                        ]
                        if len(audio_files) > 0:
                            self.handle_media2db(root)
            elif self.target_type == MmusicC.ElementType.file:
                self.handle_db2media(self.target)
            elif self.target_type == MmusicC.ElementType.folder:
                if self.result.album:
                    self.handle_db2media(self.target)
                else:
                    gen = os.walk(self.target, topdown=True)
                    for root, dirs, files in gen:
                        audio_files = [
                            file for file in files if is_supported_audio(file)
                        ]
                        if len(audio_files) > 0:
                            self.handle_db2media(root)

        elif self.source_type == MmusicC.ElementType.file:
            self.handle_files2file(self.source, self.target)

        elif self.source_type == MmusicC.ElementType.folder:
            if self.result.album:
                self.handle_album2album(self.source, self.target)
            else:
                logging.debug(
                    "Walking through tree of directory '{}'".format(self.source)
                )
                gen = os.walk(self.source, topdown=True)
                for root, dirs, files in gen:
                    root = pathlib.Path(root)
                    album_target = swap_base(self.source, root, self.target)
                    logging.debug("Current root: {}".format(root))
                    if len(dirs) == 0:
                        if self.target_type == MmusicC.ElementType.folder:
                            self.handle_album2album(root, album_target)
                        else:
                            self.handle_media2db(root)
                    elif len(files) > 0:
                        self.handle_album2album(root, album_target)
                        logging.info(
                            "Found files not in album {} at folder '{}'".format(
                                files, root
                            )
                        )

        sys.exit(0)

    def handle_files2file(self, file_source, file_target):
        if not is_supported_audio(file_source):
            logging.warning("file f'{file_source}' not a supported audio file.")
            return
        if self.get_element_type(file_target) == MmusicC.ElementType.folder:
            file_target = file_target.joinpath(file_source.stem + self.format_extension)
        if self.run_files and file_source and file_target:
            if file_target.is_file():
                logging.info("target file exists, ffmpeg skipped, returning")
                return
            if not self.result.dry_run:
                FFmpeg(
                    str(file_source),
                    str(file_target),
                    options=self.result.ffmpeg_options,
                ).run()
        if self.run_meta:
            meta_source = Metadata(file_source)
            meta_target = Metadata(file_target)
            meta_target.import_tags(
                meta_source,
                whitelist=self.whitelist,
                blacklist=self.blacklist,
                skip_none=self.result.lazy,
            )
            meta_target.write_tags(remove_existing=self.result.delete_existing_metadata)
        return

    def handle_album2album(self, album_source, album_target):
        if self.run_files and album_source and album_target:
            if not album_target.is_dir():
                album_target.mkdir(parents=True)
            for file in os.listdir(album_source):
                if is_supported_audio(file):
                    file_source = album_source.joinpath(file)
                    self.handle_files2file(file_source, album_target)
        if self.run_meta:
            if not album_target.is_dir():
                logging.warning(
                    "no target folder for given source '{}', "
                    "skipping".format(album_source)
                )
                return
            meta_source = AlbumMetadata(album_source)
            meta_target = AlbumMetadata(album_target)
            meta_target.import_tags(
                meta_source,
                whitelist=self.whitelist,
                blacklist=self.blacklist,
                skip_none=self.result.lazy,
            )
            meta_target.write_tags(remove_existing=self.result.delete_existing_metadata)

    def handle_media2db(self, album_source):
        album_target = pathlib.Path(album_source)
        if album_target.is_file():
            meta_source = Metadata(album_source)
        else:
            meta_source = AlbumMetadata(album_source)
        meta_source.export_tags_to_db()

    def handle_db2media(self, album_target):
        album_target = pathlib.Path(album_target)
        if album_target.is_file():
            meta_target = Metadata(album_target)
        else:
            meta_target = AlbumMetadata(album_target)
        meta_target.import_tags_from_db(
            whitelist=self.whitelist,
            blacklist=self.blacklist,
            skip_none=self.result.lazy,
        )
        meta_target.write_tags(remove_existing=self.result.delete_existing_metadata)

    def get_element_type(self, element):
        """Get the element type based on the pathlib.Path object.

        Instance method in clarifying that the function is instance dependent.

        """

        if not element:
            if Metadata.is_linked_database:
                return MmusicC.ElementType.database
            else:
                return MmusicC.ElementType.other

        if element.suffix in audio_loader.keys():  # target is file
            if element.exists():
                if not is_supported_audio(element):
                    self.parser.error(f"File '{element}' not supported!")
            return MmusicC.ElementType.file
        else:
            if "." in element.suffix:
                logging.info(
                    f"Path identified as folder, although it can also be an "
                    f"unsupported file with extension {element.suffix}"
                )
            return MmusicC.ElementType.folder

    class ElementType(enum.Enum):
        file = (1,)
        folder = (2,)
        database = (3,)
        other = 4


def load_tags_from_list_or_file(list_tags_or_file):
    """Load TAG-list from a list of STRING or a file.

    Where each tag (STRING) is separated by newline.

    Args:
        list_tags_or_file (list of str or str): list of tags or file_path

    Returns:
        list of str: TAGs

    """
    if len(list_tags_or_file) == 1:
        path = pathlib.Path(list_tags_or_file[0])
        if path.exists():
            with path.open(mode="r") as f:
                list_tags_or_file = [line.rstrip() for line in f]
    return get_tags_from_strs(list_tags_or_file)



import hashlib
import os
import pathlib
import time

from mmusicc import Metadata
from mmusicc.util.misc import swap_base, is_supported_audio


def cmp_files_metadata(base_a, base_b, ext_b=None):
    """return number of files with identical metadata."""
    equals = 0
    for root, dirs, files in os.walk(base_a, topdown=True):
        audio_files = [file for file in files if is_supported_audio(file)]
        for path_a in audio_files:
            path_a = pathlib.Path(root).joinpath(path_a)
            path_b = swap_base(base_a, path_a, base_b)
            if ext_b:
                path_b = path_b.with_suffix(ext_b)
            if Metadata(path_a).dict_data == Metadata(path_b).dict_data:
                equals += 1
    return equals


def save_files_hash_and_mtime(list_files, touch=False) -> dict:
    """save hashes to be used in compare_files_hash and mtime
        touches files before getting time when touch=True
    """
    if not isinstance(list_files, list):
        if os.path.isdir(list_files):
            list_files = get_file_list_tree(list_files)
        else:
            list_files = [list_files]
    hash_dict = dict()
    for file in list_files:
        file = pathlib.Path(file)
        if touch:
            file.touch()
        hash_dict[file] = (hash_file(file), file.stat().st_mtime, file.stat().st_atime)
    return hash_dict


def cmp_files_hash_and_time(list_files, hash_dict):
    """compares hashes and mtime to check if file was changed. Counts changed
        files (changed hash adds +1, changed mtime adds +100 to result)
    """
    if not isinstance(list_files, list):
        if os.path.isdir(list_files):
            list_files = get_file_list_tree(list_files)
        else:
            list_files = [list_files]
    try:
        time.sleep(0.01)
        exit_code = 0
        for file in list_files:
            file = pathlib.Path(file)
            # file has changed (content or metadata) (and was also accessed)
            if not hash_dict.get(file)[0] == hash_file(file):
                exit_code += 10000
                # the file system is not fast enough updating the metadata. Since access
                # seems to be faster and you can only modify a file by accessing it,
                # add a delay to make sure metadata gets updated.
                i = 0
                while i < 10 and hash_dict.get(file)[1] == file.stat().st_mtime:
                    time.sleep(0.02)
                    i += 1
            # file was accessed (or overwritten with original)
            if not hash_dict.get(file)[1] == file.stat().st_atime:
                exit_code += 1
            if not hash_dict.get(file)[1] == file.stat().st_mtime:
                exit_code += 100
        return exit_code
    except KeyError:
        raise Exception("hash dict not complete, key can not be checked")


def hash_file(path) -> str:
    """return sha1 hash of file"""
    # noinspection PyPep8Naming
    BUF_SIZE = 65536  # 64kb
    sha1 = hashlib.sha1()
    with open(path, "rb") as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha1.update(data)
    return sha1.hexdigest()


def get_file_list_tree(tree_root, depth=None, ret_base=False):
    """return a list off all files in folder (at given search depth)"""
    if not depth:
        depth = 100
    files = sorted(tree_root.resolve().rglob("*"))
    base_length = len(tree_root.parts)

    max_len_parts_a = base_length + depth
    for i in sorted(range(len(files)), reverse=True):
        if len(files[i].parts) > max_len_parts_a:
            files.pop(i)
        elif files[i].is_dir():
            files.pop(i)
    if ret_base:
        return files, base_length
    return files



"""

Attributes:
    mimes (set): A set of supported mime types. Empty until init.
    types (set): A set of AudioFile subclasses. Empty until init.
    loaders(dict): A dict mapping file extensions to loaders
        (function returning an AudioFile object). Empty until init.

"""


import logging
import os

from mmusicc import util

mimes = set()

loaders = {}

types = set()


def init():
    """Initialize the formats module, loading all formats for loader.

    Before this is called loading a file and unpickling will not work.
    If the module was already initialized, the initialisation is skipped.
    """

    global mimes, loaders, types

    if loaders:
        logging.debug("Formats Already Initialized, Skipping.")
        return

    base = util.get_module_dir()
    formats = util.importhelper.load_dir_modules(base, package=__package__)

    module_names = []
    for form in formats:
        name = form.__name__

        for ext in form.extensions:
            loaders[ext] = form.loader

        types.update(form.types)

        if form.extensions:
            for type_ in form.types:
                mimes.update(type_.mimes)
            module_names.append(name.split(".")[-1])

    s = ", ".join(sorted(module_names))

    logging.info("Initialized modules. Supported formats: {}.".format(s))

    if not loaders:
        raise SystemExit("No formats found!")


def get_loader(file_path):
    """Returns a callable which takes a filename and returns AudioFile

    or raises AudioFileError, or returns None.

    Args:
        file_path (str or pathlib.Path): file to get loader for.
    Returns:
        obj: callable object which can handle given file.

    """
    ext = os.path.splitext(file_path)[-1]
    return loaders.get(ext.lower())


# noinspection PyPep8Naming
def MusicFile(file_path):
    """Returns a AudioFile instance or None if file type is not supported

    Note:
        Make sure that allocationmap is initialized before loading any files.

    Args:
        file_path (pathlib.Path): path of file to load as MusicFile

    Returns:
        AudioFile: audio file instance of file in specified path
    """
    loader = get_loader(file_path)
    if loader is not None:
        return loader(file_path)
    else:
        logging.error("FileType not supported")
        raise Exception(
            "FileType '{}' not supported. Supported Loaders: {}".format(
                os.path.splitext(str(file_path))[-1], str(loaders)
            )
        )


class AudioFileError(Exception):
    """Base error for AudioFile, mostly IO/parsing related operations"""


class MutagenBug(AudioFileError):
    """Raised in is caused by a mutagen bug, so we can highlight it"""



import pathlib

from mmusicc.util.metadatadict import MetadataDict


class AudioFile:
    """Base class for all audio files wrapping tags and audio stream info.

    Attributes:
        _file       (pathlib.Path): file path of file.
        _dict_meta  (MetadataDict): metadata from file (the parsed one)
        unprocessed_tag     (dict): metadata that could'nt be associated.
            Manly used to manually update the association list with new
            tags/tag names.
    """

    def __init__(self):
        self._dict_meta = MetadataDict()
        self._file = None
        self._changed_tags = None
        self.unprocessed_tag = dict()

    @property
    def file_path(self):
        """pathlib.Path: file path of audio file."""
        return pathlib.Path(self._file.filename)

    @property
    def dict_meta(self):
        """MetadataDict: metadata of the audio file (the parsed one)."""
        return self._dict_meta

    @dict_meta.setter
    def dict_meta(self, value):
        if isinstance(value, MetadataDict):
            self._dict_meta = value
        else:
            raise TypeError("only MetadataDict allowed")

    def file_read(self):
        """reads file tags into AudioFile tag dictionary (dict_meta)."""
        raise NotImplementedError

    def file_save(self, remove_existing=False, write_empty=False):
        """saves file tags from tag dictionary (dict_meta) to AudioFile.

        Args:
            remove_existing (bool): if true clear all tags before writing.
                Defaults to False.
            write_empty     (bool): if true write tags with Empty Value, if
                false, the tag will not be created and a existing tag will be
                deleted. Behaviour might slightly differ between tag types.
                Defaults to False.
        """
        raise NotImplementedError



import logging
import logging.config


def init_logger(level=25):
    """Initialise Python Logger (function may be extended in future)"""

    if len(logging.getLogger().handlers):
        logging.debug("Logger Already Initialized, Skipping.")
        return

    # noinspection PyPep8
    config = {
        "version": 1,
        "formatters": {
            "simple": {
                "format": "%(asctime)s %(levelname)-8s %(module)-13s %(message)s",
            }
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                # "level": level if level == logging.DEBUG else logging.ERROR,
                "level": level,
                "formatter": "simple",
                "stream": "ext://sys.stdout",
            },
        },
        "root": {"level": level, "handlers": ["console"],},
    }
    logging.config.dictConfig(config)
    logging.debug("Initialized logger")
    logging.addLevelName(25, "mmusicc")



"""
Definitions for this documentation:

- TAG
    string of tag used in program (e.g. MetadataDict), also TAG display name.
- STRING(s)
    string used in the files for the TAGS. Can be a list to match
    different variants (e.g. comment, description).
- ID3
    frame names of ID3 tag standard


Attributes:
    list_tags     (list of tags): list of all recognized TAGs.
    dict_tag2id3          (dict): dict mapping TAG to ID3.
    dict_id32tag          (dict): dict mapping ID3 to TAG.
    dict_tag2strs         (dict): dict mapping TAG to STRING(s).
    dict_str2tag          (dict): dict mapping STRING to TAG.

"""


import logging
import os

import yaml

list_tui_tags = list()
list_tags = list()
dict_tag2id3 = dict()  # tag: id3
dict_id32tag = dict()  # id3: tag
dict_tag2strs = dict()  # tag: strs (list)
dict_str2tag = dict()  # str: tag
dict_auto_fill_rules = dict()


def init_allocationmap(path, force=False):
    """Initialize allocation map from config file.

    Creates dicts as lookup tables as global variables.

    Args:
        path             (str): path of yaml file containing a mapping for
            the allocation table
        force (bool, optional): if True, initializes the allocationmap even if
            it is already initialized, otherwise skips. Defaults to False.
    """

    global list_tui_tags
    global list_tags
    global dict_tag2id3
    global dict_id32tag
    global dict_tag2strs
    global dict_str2tag
    global dict_auto_fill_rules

    if len(list_tags) > 0:
        if force:
            logging.debug(
                "allocationmap Already Initialized, " "Continue (force option set)."
            )
            # reset all lists and dicts
            list_tui_tags = list()
            list_tags = list()
            dict_tag2id3 = dict()
            dict_id32tag = dict()
            dict_tag2strs = dict()
            dict_str2tag = dict()
            dict_auto_fill_rules = dict()
        else:
            logging.debug("allocationmap Already Initialized, Skipping.")
            return

    path = os.path.abspath(os.path.expanduser(path))
    if not os.path.exists(path):
        raise FileNotFoundError("config file not found '{}'".format(path))
    with open(path, "r") as f:
        dict_config = yaml.safe_load(f)
    max_pos = max(dict_config.values(), key=lambda x: int(x[0]))[0]
    if len(list_tui_tags) < max_pos:
        list_tui_tags = [None] * max_pos
    for key, value in dict_config.items():
        key = key.casefold()
        pos = value[0]
        id3_tag = value[1]
        if len(value) >= 3:
            assertions = set([v.casefold() for v in value[2]])
        else:
            assertions = set()
        if len(value) >= 4:
            if value[3]:
                if len(value[3]) == 3:
                    value[3][0] = value[3][0].casefold()
                    if not value[3][1]:
                        value[3][2] = value[3][2].casefold()
                dict_auto_fill_rules[key] = value[3]
        if len(value) >= 5:
            # placeholder for additional future
            pass

        list_tags.append(key)
        list_tui_tags[pos - 1] = key
        dict_tag2id3[key] = id3_tag
        dict_id32tag[id3_tag] = key
        strings = []
        if key not in assertions:
            strings.append(key)
        for val in assertions:
            strings.append(val)
        dict_tag2strs[key] = strings
        for string in strings:
            dict_str2tag[string] = key

    i = len(list_tui_tags) - 1
    while list_tui_tags[i] is None:
        list_tui_tags.pop(i)
        i -= 1

    logging.debug("allocationmap Initialized, found tags: {}".format(list_tags))


def get_tags_from_strs(tags):
    """Convert a list of STRING(s) to list of TAGs.

    Args:
        tags (list of str): STRING(s).

    Returns:
        list of str: TAGs.
    """
    ret_tag = list()
    for string in tags:
        try:
            ret_tag.append(dict_str2tag[string])
        except KeyError:
            raise KeyError(
                "String '{}' could not be associated with any tag.".format(string)
            )
    return ret_tag



import pathlib

from mmusicc.util.misc import get_the_right_one, swap_base, process_white_and_blacklist


def test_black_and_whitelist():
    whitelist = ["a", "b", "c", "d", "e", "f"]
    blacklist = ["e", "f", "g", "h", "a"]
    result = process_white_and_blacklist(whitelist, blacklist)
    assert result == ["b", "c", "d"]


def test_get_the_right_one():
    plist = [
        "a/b/c/fuu/beam/hello.mp3",
        "d/e/f/fuu/bar/world.mp3",
        "g/h/foo/bar/hello.mp3",
        pathlib.Path("l/fee/bar/world.mp3"),
    ]
    match = "m/n/foo/bar/hello.mp3"

    the_right_one = get_the_right_one(plist, match)

    assert str(the_right_one) == plist[2]


def test_swap_base():
    root_a = pathlib.Path("base/A")
    path_a = pathlib.Path("base/A/fuu/bar/")
    root_b = pathlib.Path("esab/B")
    path_b_expected = pathlib.Path("esab/B/fuu/bar/")
    assert swap_base(root_a, path_a, root_b) == path_b_expected



import mutagen

import mmusicc.util.allocationmap as am
from mmusicc.formats._audio import AudioFile
from mmusicc.util.metadatadict import Empty, scan_dictionary, AlbumArt
from mmusicc.util.util import text_parser_get, join_str_list

extensions = [".mp3", ".mp2", ".mp1", ".mpg", ".mpeg"]
"""list of all extensions associated with this module"""
# loader   see bottom
# types    see bottom


def _to_album_art(frame):
    album_art = AlbumArt()
    album_art.desc = frame.desc
    album_art.ptype = frame.type
    album_art.data = frame.data
    album_art.mime = frame.mime
    return album_art


class MP3File(AudioFile):
    """Tag object for mp3 files using ID3 tag standard.

    Instance created by loader (at the bottom of this code) which in this case
    is this class itself (vgl. xiph).

    Args:
        file_path      (str): file path of represented audio file
    """

    format = "MPEG-1/2"
    mimes = ["audio/mp3", "audio/x-mp3", "audio/mpeg", "audio/mpg", "audio/x-mpeg"]

    def __init__(self, file_path):
        super().__init__()
        self._file = mutagen.File(file_path)
        self._changed_tags = None

    def file_read(self, drop_role=True):
        """reads file tags into AudioFile tag dictionary (dict_meta).

        First tries to associate ID3 tags, than takes all txxx tags and runs
        them through the scan dictionary function.

        Args:
            drop_role (bool): if True, drop the role of paired text frame and
                save as a list of strings. If False, save the
        """
        tags_txxx = dict()

        for frame in self._file.values():

            frame_type = type(frame)
            frame_type_parent = frame_type.mro()[1]

            # handle each frame depending on parent class
            if frame.FrameID == "APIC":
                tag_val = _to_album_art(frame)
            elif frame_type is mutagen.id3.TXXX:
                tags_txxx[frame.desc] = frame.text[0]
                continue
            elif frame_type_parent is mutagen.id3.TextFrame:
                tag_val = frame.text[0]
            elif frame_type_parent is mutagen.id3.PairedTextFrame:
                tmp_key = "PairedTextFrame not supported:" + frame.HashKey
                self.unprocessed_tag[tmp_key] = frame.people
                continue
            elif frame_type_parent is mutagen.id3.NumericPartTextFrame:
                tag_val = frame.text[0]
            elif frame_type_parent is mutagen.id3.NumericTextFrame:
                tag_val = frame.text[0]
            elif frame_type_parent is mutagen.id3.TimeStampTextFrame:
                tag_val = frame.text[0].get_text()
            elif frame_type_parent is mutagen.id3.Frame:
                tag_val = frame.text
            else:
                raise Exception("frame not implemented %s", frame.HashKey)

            if Empty.is_empty(tag_val):
                tag_val = Empty()

            tag_key = am.dict_id32tag.get(frame.FrameID)
            if tag_key:
                if isinstance(tag_val, Empty) or isinstance(tag_val, AlbumArt):
                    val = tag_val
                else:
                    val = text_parser_get(tag_val)
                    if len(val) == 1:
                        val = val[0]
                self.dict_meta[tag_key] = val
            else:
                self.unprocessed_tag[frame.HashKey] = tag_val

        if len(tags_txxx) > 0:
            self.unprocessed_tag.update(scan_dictionary(tags_txxx, self.dict_meta))

    def file_save(self, remove_existing=False, write_empty=False, remove_v1=False):
        """saves file tags from tag dictionary (dict_meta) to AudioFile.

        Note:
            write_empty may have no effect. Since mutagen will not load empty
            tags it can't be checked. Correction or Info is appreciated.

        Args:
            remove_existing ('bool', optional): if true clear all tags on file
                before writing. Defaults to False.
            write_empty     (bool): Only affects TXXX tags. Existing tags will
                always be set to None. If true create empty TXXX tags with
                value none. If false no tag will be created or existing
                TXXX tag on file will be deleted. Defaults to False.
            remove_v1       ('bool'): If True, remove existing ID3.V1 tags.
                Defaults to False.
        """
        self._changed_tags = list()

        audio = self._file
        if audio.tags is None:
            audio.add_tags()

        if remove_existing:
            audio.delete()

        if audio.tags is None:
            audio.add_tags()

        new_tags = self.dict_meta.copy_not_none()
        tag_remove = list()
        for tag_key, value in new_tags.items():

            id3_tag = am.dict_tag2id3.get(tag_key)
            # load existing frame (or not)
            if id3_tag == "TXXX":
                txxx_tags = audio.tags.getall("TXXX")
                for tx in txxx_tags:
                    if tag_key in tx.desc.casefold():
                        frame = [tx]
                        break
                else:
                    frame = []
            else:
                frame = audio.tags.getall(id3_tag)
            if frame:
                if len(frame) > 1:
                    raise Exception(f"to many tags of {id3_tag} found")
                else:
                    frame = frame[0]
                # if it exists, check if it has to be deleted
                if Empty.is_empty(value) and not write_empty:
                    tag_remove.append(tag_key)
                    audio.tags.delall(frame.FrameID)
                    continue
            else:
                # if it does not exist, create a new one and add it to file
                frame = eval("mutagen.id3.{}()".format(id3_tag))
                audio.tags.add(frame)

            if Empty.is_empty(value):
                value = ""

            frame_type = type(frame)
            frame_type_parent = frame_type.mro()[1]

            if frame_type is mutagen.id3.TXXX:
                if tag_key not in frame.desc.casefold():
                    self._set_value(frame, "desc", tag_key)

            if frame.FrameID == "APIC":
                self._fill_apic_frame_with_albumart(frame, value)
            elif frame_type_parent is mutagen.id3.PairedTextFrame:
                # text frame support dropped
                continue
            else:
                if isinstance(value, list):
                    value = join_str_list(value)
                self._set_value(frame, "text", value)
            if not frame.encoding:
                self._set_value(frame, "encoding", mutagen.id3.Encoding.UTF16)
            try:
                lang = frame.lang
                if not lang:
                    self._set_value(frame, "lang", "eng")
            except AttributeError:
                pass

        if remove_v1:
            v1 = 0  # ID3v1 tags will be removed
        else:  # 1  # ID3v1 tags will be updated  but not added
            v1 = 2  # ID3v1 tags will be created and / or updated

        if len(self._changed_tags) == 0:
            if remove_existing and len(self.unprocessed_tag):
                pass
            else:
                return

        audio.save(v1=v1, v2_version=4, v23_sep=None)

    def _fill_apic_frame_with_albumart(self, frame, album_art):
        self._set_value(frame, "desc", album_art.desc)
        self._set_value(frame, "type", album_art.ptype)
        self._set_value(frame, "data", album_art.data)
        self._set_value(frame, "mime", album_art.mime)

    def _set_value(self, frame, attr, value):
        org = getattr(frame, attr)
        setattr(frame, attr, value)
        new = getattr(frame, attr)
        if not org == new:
            self._changed_tags.append((frame, attr, org, new))


types = [MP3File]
"""list of all subclasses of AudioFile in this module"""

loader = MP3File
"""class alias for dynamic loading of available AudioFile subclasses"""



#  Copyright (c) 2020 Johannes Nolte
#  SPDX-License-Identifier: GPL-3.0-or-later

import base64

import mutagen
from mutagen.flac import FLAC, Picture
from mutagen.oggvorbis import OggVorbis

from mmusicc.formats._audio import AudioFile
from mmusicc.formats._misc import AudioFileError
from mmusicc.util.metadatadict import Empty, scan_dictionary, AlbumArt

extensions = [".ogg", ".oga", ".flac"]
"""list of all extensions associated with this module"""
# loader   see bottom
# types    see bottom
ogg_formats = [FLAC, OggVorbis]


def _to_album_art(picture_or_data):
    if isinstance(picture_or_data, str):
        picture = Picture(base64.b64decode(picture_or_data))
    else:
        picture = picture_or_data
    album_art = AlbumArt()
    album_art.desc = picture.desc
    album_art.ptype = picture.type
    album_art.data = picture.data
    album_art.mime = picture.mime
    return album_art


def _album_art_to_picture(album_art):
    pic = Picture()
    pic.desc = album_art.desc
    pic.type = album_art.ptype
    pic.data = album_art.data
    pic.mime = album_art.mime
    return pic


def _album_art_to_picture_metablock(album_art):
    return base64.b64encode(_album_art_to_picture(album_art).write()).decode("ascii")


class VCFile(AudioFile):
    """Tag object for all files using tags of the xiph v-comment tag standard.

    Instance created by loader (at the bottom of this code).

    Args:
        audio (mutagen.File): file as mutagen file object.
    """

    format = "Unknown Mutagen + vorbiscomment"
    MutagenType = None

    can_change_images = True

    def __init__(self, audio):
        super().__init__()
        self._file = audio
        self._changed_tags = None

    def file_read(self):
        """reads file tags into AudioFile tag dictionary (dict_meta)."""
        if self._file.tags:
            tags = self._file.tags.as_dict()

            pictures = list()
            for data in tags.pop("metadata_block_picture", []):
                pictures.append(_to_album_art(data))
            try:
                for pic in self._file.pictures:
                    pictures.append(_to_album_art(pic))
            except AttributeError:
                pass

            if pictures:
                tags.update({"albumart": pictures})

            for key in tags.keys():
                if len(tags[key]) == 1:
                    tags[key] = tags[key][0]
            self.unprocessed_tag.update(scan_dictionary(tags, self._dict_meta))

    def file_save(self, remove_existing=False, write_empty=False):
        """saves file tags from tag dictionary (dict_meta) to AudioFile.

            if no value changes the file is not saved, therefore there will be
            no changes at file (mtime does not change). When remove_existing
            is set, the file is saved, when unprocessed tags exists.

        Args:
            remove_existing (bool): if true clear all tags on file before
                writing. Defaults to False.
            write_empty     (bool): if true write empty tags an therefore
                create a tag with content "". If false no tag will be created
                and existing tags on file that would be overwritten with ""
                will be deleted. Defaults to False.

        Note:
            A tag not existing in dict_meta or being None in dict_meta will not be
            deleted and es kept unchanged. The (possible empty) tag will be deleted,
            when an Empty value is in source and write_empty is False.
        """
        self._changed_tags = list()

        audio = self._file
        if audio.tags is None:
            audio.add_tags()

        meta = self.dict_meta.copy_not_none()

        # handle write empty policy
        for key, value in list(meta.items()):
            if Empty.is_empty(value):
                if write_empty:
                    meta[key] = ""
                else:
                    self._changed_tags.append((key, audio[key], "*"))
                    del meta[key]
                    del audio[key]

        cover_keys = list()

        for key in meta.keys():
            tag = meta.get(key)
            if isinstance(tag, AlbumArt):
                cover_keys.append(key)
                continue
            self._set_value(audio, key, tag)

        new_art = [meta.get(key) for key in cover_keys]
        if hasattr(audio, "pictures"):
            org_art = [_to_album_art(cover) for cover in audio.pictures]
            clear = False
            for art in org_art.copy():
                if art not in new_art:
                    self._changed_tags.append(("albumart", art, "*"))
                    org_art.remove(art)
                    clear = True
            if clear:
                # clear all
                audio.clear_pictures()
                # add what was cleared to much
                [audio.add_picture(_album_art_to_picture(art)) for art in org_art]
            for art in new_art:
                if art not in org_art:
                    self._changed_tags.append(("albumart", "*", art))
                    audio.add_picture(_album_art_to_picture(art))
        else:
            org_art = dict()
            for block in audio.get("metadata_block_picture", []):
                org_art[_to_album_art(block)] = block

            for art in org_art.keys():
                if art not in new_art:
                    self._changed_tags.append(("albumart", art, "*"))
                    audio["metadata_block_picture"].remove(org_art.get(art))
            for art in new_art:
                if art not in org_art:
                    self._changed_tags.append(("albumart", "*", art))
                    if "metadata_block_picture" not in audio:
                        audio[
                            "metadata_block_picture"
                        ] = _album_art_to_picture_metablock(art)
                    else:
                        audio["metadata_block_picture"].append(
                            _album_art_to_picture_metablock(art)
                        )

            if audio.get("metadata_block_picture") is None:
                pass
            elif len(audio.get("metadata_block_picture", [])) > 0:
                meta["metadata_block_picture"] = "to not delete"
            else:
                del audio["metadata_block_picture"]

        if remove_existing:
            for key in [k for k in audio if k not in meta]:
                self._changed_tags.append((key, audio[key], "*"))
                del audio[key]

        if len(self._changed_tags) == 0:
            return

        # self._file.update(meta)
        self._file.save()

    def _set_value(self, dictionary, key, value):
        try:
            org = dictionary[key]
            dictionary[key] = value
            new = dictionary[key]
            if not new == org:
                self._changed_tags.append((key, org, new))
        except KeyError:
            dictionary[key] = value
            self._changed_tags.append((key, "*", value))


class OggFile(VCFile):
    """File type specific subclass of VCFile"""

    format = "Ogg Vorbis"
    mimes = ["audio/vorbis", "audio/ogg", "audio/x-ogg"]
    MutagenType = OggVorbis


class FLACFile(VCFile):
    """File type specific subclass of VCFile"""

    format = "FLAC"
    mimes = ["audio/flac", "audio/x-flac"]
    MutagenType = FLAC


# list of subclasses of Audio File.
# Dynamically fills the supported type list with the above defined classes.
types = []
"""list of all subclasses of AudioFile in this module"""
for var in list(globals().values()):
    if getattr(var, "MutagenType", None):
        types.append(var)


def loader(file_path):
    """loads the given file into a VCFile object and returns it.

    Args:
        file_path (str): path of file to be loaded.
    Returns:
        AudioFile: VCFile tag objects with given file loaded.
    Raises:
        AudioFileError: if file type could not be determined.
    """
    audio = mutagen.File(file_path, options=ogg_formats)
    if audio is None:
        raise AudioFileError("file type could not be determined")
    mutagen_type = type(audio)
    for xiph in [FLACFile, OggFile]:
        if mutagen_type is getattr(xiph, "MutagenType", None):
            # initialize a VCFile Instance and return it
            return xiph(audio)
    raise AudioFileError("file type could not be determined")



#  Copyright (c) 2020 Johannes Nolte
#  SPDX-License-Identifier: GPL-3.0-or-later

import setuptools


with open("README.rst", "r") as fh:
    long_description = fh.read()

version = {}
with open("./mmusicc/version.py") as fp:
    exec(fp.read(), version)

setuptools.setup(
    name="mmusicc",
    version=version["__version__"],
    author="Johannes Nolte",
    author_email="jonolt@mailbox.org",
    description=(
        "A lightweight audio file and metadata control and synchronization "
        "program to transfer the changes made in a master music library to a "
        "derived music library."
    ),
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/jonolt/mmusicc",
    project_urls={
            "Documentation": "https://github.com/jonolt/mmusicc",
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: POSIX",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Multimedia :: Sound/Audio :: Conversion",
        "Topic :: Multimedia :: Sound/Audio :: Editors",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
    ],
    python_requires=">=3.6",
    package_data={"mmusicc": ["data/config.yaml"]},
    install_requires=["mutagen>=1.41.0", "PyYAML>=3.01", "SQLAlchemy>=1.3.0"],
    entry_points={"console_scripts": ["mmusicc = mmusicc.__main__:main"]},
)



import pytest

from mmusicc.metadata import GroupMetadata, AlbumMetadata, Div, Empty
from ._util import *

_dict_files = {
    Metadata: "1-str_title_A.flac",
    GroupMetadata: ["1-str_title_A.flac", "2-str_title_B.flac", "3-str_title_C.flac"],
    AlbumMetadata: None,  # _create_file_path return an folder path
}


def test_dummy_for_init(allocation_map, audio_loaders):
    """test not needed here but otherwise fixture (module level) will not be
        loaded and the init functions are not called
    """
    assert len(allocation_map.list_tags) == 16
    assert len(audio_loaders) > 0


@pytest.fixture
def modified_metadata(request, dir_lib_x_flac):
    """provides a Metadata object loaded from the flac source with modified
        values for test.
    """
    if request.param == AlbumMetadata or request.param == GroupMetadata:
        metadata_path = _create_file_path(GroupMetadata, dir_lib_x_flac)
        meta = GroupMetadata(metadata_path)
        meta.reset_meta()
        meta.set_tag("album", "fuubar")
        meta.set_tag("artist", "quodlibet")
    else:
        metadata_path = _create_file_path(Metadata, dir_lib_x_flac)
        meta = Metadata(metadata_path)
        meta.dict_data.reset()
        meta.set_tag("album", "fuubar")
        meta.set_tag("artist", "quodlibet")
    return request.param, meta


def test_database_link(dir_lib_x_flac, path_database):
    """test linking and unlinking the database and corresponding properties."""
    metadata_path = _create_file_path(Metadata, dir_lib_x_flac)
    meta = Metadata(metadata_path)
    with pytest.raises(Exception):
        meta.export_tags_to_db()
    with pytest.raises(Exception):
        Metadata.unlink_database()
    Metadata.link_database(str(path_database))
    assert Metadata.is_linked_database
    with pytest.raises(Exception):
        Metadata.link_database(str(path_database))
    Metadata.unlink_database()
    assert not Metadata.is_linked_database
    Metadata.link_database(str(path_database))


@pytest.mark.parametrize("class_meta", _dict_files.keys())
def test_read_tags(class_meta, dir_lib_x_flac):
    metadata_path = _create_file_path(class_meta, dir_lib_x_flac)
    meta = class_meta(metadata_path)
    assert isinstance(meta, class_meta)
    assert meta._dict_data.get("album")
    if class_meta is Metadata:
        assert meta.file_path
    else:  # Group and Album are basically the same
        assert isinstance(meta._dict_data.get("title"), Div)


@pytest.mark.parametrize("class_meta", _dict_files.keys())
@pytest.mark.parametrize("rem_ex, write_empty", [(False, False), (True, True)])
def test_write_tags(class_meta, dir_lib_x_flac, rem_ex, write_empty):
    metadata_path = _create_file_path(class_meta, dir_lib_x_flac)
    meta = class_meta(metadata_path)
    meta.set_tag("album", "fuubar")
    meta.set_tag("albumartist", Empty())
    meta.set_tag("artist", None)
    meta.write_tags(remove_existing=rem_ex, write_empty=write_empty)
    meta_2 = class_meta(metadata_path)
    assert meta_2._dict_data.get("album") == "fuubar"

    if rem_ex:
        assert meta_2._dict_data.get("artist") is None
    else:
        assert meta_2._dict_data.get("artist") == "str_artist"

    if write_empty:
        assert Empty.is_empty(meta_2._dict_data.get("albumartist"))
    else:
        assert meta_2._dict_data.get("albumartist") is None


@pytest.mark.parametrize("skip_none", [True, False])
@pytest.mark.parametrize("modified_metadata", _dict_files.keys(), indirect=True)
def test_import_tags(modified_metadata, dir_lib_x_flac, skip_none):
    class_meta, import_source = modified_metadata
    metadata_path = _create_file_path(class_meta, dir_lib_x_flac)
    meta = class_meta(metadata_path)
    meta.import_tags(import_source, whitelist=["album", "title"], skip_none=skip_none)
    # Assert proper import
    assert meta.get_tag("album") == "fuubar"
    # Assert whitelist (blacklist does not need to be tested here, since the
    # actual whitelist of the called func is computed from black and whitelist.
    assert meta.get_tag("artist") == "str_artist"
    # Assert skip option (parametrized)
    if skip_none:
        if class_meta is Metadata:
            assert meta.get_tag("title") == "str_title_A"
        else:
            assert isinstance(meta.get_tag("title"), Div)
    else:
        assert meta.get_tag("title") is None


@pytest.mark.parametrize("skip_none", [True, False])
@pytest.mark.parametrize("class_meta", _dict_files.keys())
def test_import_db_tag(class_meta, dir_lib_x_flac, path_database, skip_none):
    # artist: assert whitelist
    # album : assert correct import
    # title : assert skip none
    if Metadata.is_linked_database:
        Metadata.unlink_database()
    Metadata.link_database(str(path_database))
    metadata_path = _create_file_path(class_meta, dir_lib_x_flac)
    meta = class_meta(metadata_path)
    meta.set_tag("album", "fuu")
    meta.set_tag("artist", "bar")
    meta.import_tags_from_db(whitelist=["album", "title"], skip_none=skip_none)
    assert meta.get_tag("artist") == "bar"
    if skip_none:
        if class_meta is Metadata:
            assert meta.get_tag("title") == "str_title_A"
        else:  # Group and Album are basically the same
            assert isinstance(meta.get_tag("title"), Div)
    else:
        assert meta.get_tag("title") is None


@pytest.mark.parametrize("class_meta", _dict_files.keys())
def test_export_db_tag(class_meta, dir_lib_x_flac, temp_database):
    # artist: assert whitelist
    # album : assert correct import
    if Metadata.is_linked_database:
        Metadata.unlink_database()
    Metadata.link_database(str(temp_database))
    metadata_path = _create_file_path(class_meta, dir_lib_x_flac)
    meta = class_meta(metadata_path)
    meta.export_tags_to_db()
    meta._dict_data.reset()
    meta.import_tags_from_db()
    assert meta.get_tag("artist") == "str_artist"
    if class_meta is Metadata:
        assert meta.get_tag("title") == "str_title_A"
    else:  # Group and Album are basically the same
        assert isinstance(meta.get_tag("title"), Div)


def test_load_empty_meta(dir_lib_x_flac):
    meta = Metadata()
    assert not meta.audio_file_linked
    meta.link_audio_file(str(dir_lib_x_flac.joinpath("2-str_title_B.flac")))
    assert isinstance(meta, Metadata)
    assert meta.audio_file_linked
    assert not meta.get_tag("album")
    meta.read_tags()
    assert meta.get_tag("album")


def test_dry_run(dir_lib_x_flac, temp_database):
    Metadata.dry_run = True
    if Metadata.is_linked_database:
        Metadata.unlink_database()
    Metadata.link_database(str(temp_database))
    metadata_path = _create_file_path(Metadata, dir_lib_x_flac)
    th = save_files_hash_and_mtime(metadata_path)
    meta = Metadata(metadata_path)
    meta.set_tag("album", "fuubar")
    meta.write_tags()
    meta.export_tags_to_db()
    meta.read_tags()
    assert meta.get_tag("album") != "fuubar"
    with pytest.raises(KeyError):
        meta.import_tags_from_db()
    # reset class variable to default
    assert cmp_files_hash_and_time(metadata_path, th) == 1
    Metadata.dry_run = False


@pytest.mark.skip(reason="not implemented - not so important")
def test_div_object():
    pass


def _create_file_path(class_meta, dir_lib_a_flac):
    files = _dict_files.get(class_meta)
    if not files:
        return str(dir_lib_a_flac)
    elif isinstance(files, str):
        return str(dir_lib_a_flac.joinpath(files))
    else:
        return [str(dir_lib_a_flac.joinpath(f)) for f in files]



from distutils.dir_util import copy_tree
from distutils.file_util import copy_file

import pytest

from mmusicc import MmusicC
from mmusicc.util.ffmpeg import FFRuntimeError
from ._util import *


class Ste:
    """Source Target Expected. Object holding files path for testing"""

    def __init__(self, combo, path_s, path_t, path_e, add_args=None):
        self.combo = combo
        self.path_s = path_s  # input file (lib A)
        self.path_t = path_t  # output/changed file (working copy)
        self.path_e = path_e  # expected result (lib B)
        self.add_args = add_args  # additional arguments for some tests


@pytest.fixture
def ste(request, dir_lib_a_flac, dir_lib_test, dir_lib_b_ogg, dir_lib_c_ogg):
    #            session         function      session
    """provide Paths (source target expected) for the different scenarios"""
    combo = request.param
    if combo == "file-->file":
        return Ste(
            combo,
            dir_lib_a_flac.joinpath(
                "artist_quodlibet/album_bar_-_single_(2020)/01_track1.flac"
            ),
            dir_lib_test.joinpath("01_track1.ogg"),
            dir_lib_b_ogg.joinpath(
                "artist_quodlibet/album_bar_-_single_(2020)/01_track1.ogg"
            ),
        )
    elif combo == "file-->folder":
        return Ste(
            combo,
            dir_lib_a_flac.joinpath(
                "artist_quodlibet/album_bar_-_single_(2020)/01_track1.flac"
            ),
            dir_lib_test,
            dir_lib_b_ogg.joinpath(
                "artist_quodlibet/album_bar_-_single_(2020)/01_track1.ogg"
            ),
        )
    elif combo == "folder-->folder":
        return Ste(combo, dir_lib_a_flac, dir_lib_test, dir_lib_b_ogg)
    elif combo == "album-->album":
        return Ste(
            combo,
            dir_lib_a_flac.joinpath("artist_puddletag"),
            dir_lib_test,
            dir_lib_b_ogg.joinpath("artist_puddletag"),
            "--album",
        )
    elif combo == "folder-->folder_part":
        # folder with missing files and wrong mata to be updated
        copy_tree(str(dir_lib_c_ogg), str(dir_lib_test))
        return Ste(combo, dir_lib_a_flac, dir_lib_test, dir_lib_b_ogg)
    else:
        raise ValueError("combo does not exists")


class TestMetadataOnly:
    @pytest.mark.parametrize("combo", ["file-->file", "file-->folder"])
    @pytest.mark.parametrize(
        "opt",
        [
            None,
            "--lazy",
            "--delete-existing-metadata",
            "--lazy --delete-existing-metadata",
        ],
    )
    def test_file_file(self, dir_lib_a_flac, dir_lib_c_ogg, dir_lib_test, opt, combo):
        """test file to file metadata sync with different options"""
        path_copy_source = dir_lib_c_ogg.joinpath(
            "artist_quodlibet/album_bar_-_single_(2020)/01_track1.ogg"
        )
        path_s = dir_lib_a_flac.joinpath(
            "artist_quodlibet/album_bar_-_single_(2020)/01_track1.flac"
        )
        path_t = dir_lib_test.joinpath("01_track1.ogg")
        copy_file(str(path_copy_source), str(path_t))

        org_file_list = [path_t]
        saved_file_info = save_files_hash_and_mtime(org_file_list, touch=True)

        if combo == "file-->folder":
            path_t2 = dir_lib_test
            opt_f = [opt, "-f ogg"]
        else:
            opt_f = opt
            path_t2 = path_t

        _assert_run_mmusicc(
            "--only-meta", "--source", path_s, "--target", path_t2, opt_f
        )

        assert cmp_files_hash_and_time(org_file_list, saved_file_info) > 0
        metadata = Metadata(str(path_t))
        assert metadata.get_tag("album") == "Bar - Single"
        assert metadata.get_tag("date") == "2020"
        assert metadata.get_tag("artist") == "Quod Libet"
        # assert metadata.get_tag("composer") == "should not be here"

        if opt and "--delete-existing-metadata" in opt:
            assert len(metadata.unprocessed_tag) == 0
            if "--lazy" in opt:
                assert metadata.get_tag("composer") == "should not be here"
            else:
                assert metadata.get_tag("composer") is None
        else:
            assert len(metadata.unprocessed_tag) > 0
            assert metadata.get_tag("composer") == "should not be here"

    def test_folder_folder(self, dir_lib_a_flac, dir_lib_b_ogg, dir_lib_test):
        """test folder folder metadata sync"""
        copy_tree(str(dir_lib_b_ogg), str(dir_lib_test))
        saved_file_info = save_files_hash_and_mtime(dir_lib_test, touch=True)
        _assert_run_mmusicc(
            "--only-meta",
            "--source",
            dir_lib_a_flac,
            "--target",
            dir_lib_test,
            "-f .ogg",
        )
        # check no file was modified (11 files were accessed: 11*100=1000)
        assert cmp_files_hash_and_time(dir_lib_test, saved_file_info) == 11

    @pytest.mark.parametrize("opt", [None, "--lazy", "--delete-existing-metadata"])
    def test_folder_folder_part(self, dir_lib_a_flac, dir_lib_c_ogg, dir_lib_test, opt):
        """test folder folder metadata sync, where target has not got all
            elements of source folder
        """
        copy_tree(str(dir_lib_c_ogg), str(dir_lib_test))
        saved_file_info = save_files_hash_and_mtime(dir_lib_test, touch=True)
        _assert_run_mmusicc(
            "--only-meta",
            "--source",
            dir_lib_a_flac,
            "--target",
            dir_lib_test,
            "-f .ogg",
            opt,
        )

        equal_to_lib_b = cmp_files_metadata(dir_lib_test, dir_lib_a_flac, ext_b=".flac")
        if opt is None:
            # 2 files CD_02 got empty tags not present in source which are not deleted
            # by the standard operation. At Single the composer is imported as None and
            # therefore unchanged on file, while some wrong tags are changed. (7-2-1=4)
            assert equal_to_lib_b == 4
        elif "--lazy" in opt:
            # at import the the composer tag is not overwritten with None from source
            # also the empty values in CD_02 are not replaced with none. Since
            # write_empty is by default False, a value that is Empty in Metadata will
            # be deleted on file, therefore CD_02 has no Empty tags left.
            assert equal_to_lib_b == 6
        else:
            # only metadata existing in A is left since org data is deleted
            assert equal_to_lib_b == 7

        # check no file but 3 were modified
        # 7 files were accessed, 3 modified: 7*100+3=703)
        cmp_th = cmp_files_hash_and_time(dir_lib_test, saved_file_info)
        if cmp_th == 30207 or cmp_th == 10007:
            pytest.xfail(
                "strange behaviour that occurs now and then and "
                "could be explained yet. Since the hash has changed "
                "the file must have been modified"
            )

        if opt is None:
            # Since the files in CD_02 are unchanged only one file is modified.
            assert cmp_th == 10107
        else:
            assert cmp_th == 30307

    def test_white_and_blacklist(
        self, dir_lib_a_flac, dir_lib_c_ogg, dir_lib_test, dir_subpackages
    ):
        path_copy_source = dir_lib_c_ogg.joinpath(
            "various_artists/album_best_hits_compilation_(2010)/" "CD_01/01_track1.ogg"
        )
        path_whitelist = dir_subpackages.joinpath("whitelist.txt")
        list_blacklist = ["artist"]
        path_s = dir_lib_a_flac.joinpath(
            "various_artists/album_best_hits_compilation_(2010)/" "CD_02/02_track2.flac"
        )
        path_t = dir_lib_test.joinpath("01_track1.ogg")
        copy_file(str(path_copy_source), str(path_t))
        _assert_run_mmusicc(
            "--only-meta",
            "--source",
            path_s,
            "--target",
            path_t,
            "--white-list-tags",
            path_whitelist,
            "--black-list-tags",
            list_blacklist,
        )
        metadata = Metadata(str(path_t))
        assert metadata.get_tag("album") == "best hists compilation"  # white
        assert metadata.get_tag("title") == "track2"  # white
        assert metadata.get_tag("discnumber") == "1" or "01"  # black
        assert metadata.get_tag("tracknumber") == "2" or "02"  # black
        assert metadata.get_tag("artist") == "hello"  # black

    def test_file_database_in_and_export(
        self, dir_lib_a_flac, dir_lib_c_ogg, dir_lib_test
    ):
        database_path = dir_lib_test.joinpath("fuubar.db3")
        path_s = dir_lib_a_flac.joinpath(
            "artist_quodlibet/album_bar_-_single_(2020)/01_track1.flac"
        )
        path_copy_s = dir_lib_c_ogg.joinpath(
            "artist_quodlibet/album_bar_-_single_(2020)/01_track1.ogg"
        )
        path_t = dir_lib_test.joinpath("01_track1.ogg")
        _assert_run_mmusicc(
            "--source", path_s, "--target-db", database_path,
        )
        assert pathlib.Path(database_path).is_file()

        copy_file(str(path_copy_s), str(dir_lib_test))
        saved_file_info = save_files_hash_and_mtime(dir_lib_test, touch=True)
        assert not Metadata(path_s).dict_data == Metadata(path_t).dict_data

        _assert_run_mmusicc(
            "--source-db",
            database_path,
            "--target",
            path_t,
            "--delete-existing-metadata",
        )
        # one changed audio file + one database
        assert cmp_files_hash_and_time(dir_lib_test, saved_file_info) == 10102
        assert Metadata(path_s).dict_data == Metadata(path_t).dict_data

    def test_folder_database_in_and_export(
        self, dir_lib_a_flac, dir_lib_b_ogg, dir_lib_c_ogg, dir_lib_test,
    ):
        database_path = dir_lib_test.joinpath("fuubar.db3")
        _assert_run_mmusicc(
            "--source", dir_lib_a_flac, "--target-db", database_path,
        )
        assert pathlib.Path(database_path).is_file()

        copy_tree(str(dir_lib_c_ogg), str(dir_lib_test))
        copy_tree(str(dir_lib_b_ogg), str(dir_lib_test), update=True)
        saved_file_info = save_files_hash_and_mtime(dir_lib_test, touch=True)
        _assert_run_mmusicc(
            "--target",
            dir_lib_test,
            "--source-db",
            database_path,
            "-f .ogg",
            "--delete-existing-metadata",
        )
        # 11 audio files + 1 database file. 3 audio are changed.
        assert cmp_files_hash_and_time(dir_lib_test, saved_file_info) == 30312
        assert cmp_files_metadata(dir_lib_b_ogg, dir_lib_test) == 11


@pytest.mark.parametrize("ste", ["file-->file"], indirect=True)
class TestConversionFileFile:
    """Test convert file to file operation and options"""

    @pytest.mark.parametrize("opt_format", ["mp3", "ogg"])
    def test_option_format(self, ste, opt_format):
        """ both cases are converted to mp3 since -f is ignored and the
            extension is known from the target
        """
        # not tested if format can be neglected at file-->file
        _assert_run_mmusicc(
            "--only-files", "-s", ste.path_s, "-t", ste.path_t, "-f", opt_format
        )

        assert pathlib.Path(ste.path_t).is_file()

    def test_ffmpeg_options(self, ste):
        """test if ffmpeg options are passed through"""
        # byte objects are not splited but converted to string
        _assert_run_mmusicc(
            "--only-files",
            "-s",
            ste.path_s,
            "-t",
            ste.path_t,
            "--ffmpeg-options",
            b"-q:a 9",
            "-f",
            ".ogg",
        )
        assert pathlib.Path(ste.path_t).is_file()
        # org file size is > 50000
        assert pathlib.Path(ste.path_t).stat().st_size < 10000

    def test_catch_ffmpeg_error(self, ste):
        """test if ffmpeg options are passed through and a exception is raised
            for a invalid option string
        """
        with pytest.raises(FFRuntimeError):
            # byte objects are not splited but converted to string
            _assert_run_mmusicc(
                "--only-files",
                "-s",
                ste.path_s,
                "-t",
                ste.path_t,
                "--ffmpeg-options",
                b"-q: fuubar",
            )


@pytest.mark.parametrize("ste", ["file-->folder"], indirect=True)
class TestConversionFileFolder:
    """Test convert file to folder operation and options"""

    @pytest.mark.parametrize("opt_format", ["mp3", "ogg", ".ogg"])
    def test_option_format(self, ste, opt_format):
        # both extension with and without leading dot can be used
        _assert_run_mmusicc(
            "--only-files", "-s", ste.path_s, "-t", ste.path_t, "-f", opt_format
        )
        if not opt_format == ".ogg":
            opt_format = "." + opt_format
        target_file = ste.path_s.stem + opt_format
        assert pathlib.Path(ste.path_t).joinpath(target_file).is_file()


logic_params = ["folder-->folder", "album-->album", "folder-->folder_part"]


@pytest.mark.parametrize("ste", logic_params, indirect=True)
class TestConversionFolderFolder:
    """Test convert folder to folder operation and options"""

    def test_tree_logic(self, ste):
        """test if the tree in target folder is created correctly"""
        org_file_list = get_file_list_tree(ste.path_t)
        saved_file_info = save_files_hash_and_mtime(org_file_list, touch=True)
        _assert_run_mmusicc(
            "--only-files",
            "-s",
            ste.path_s,
            "-t",
            ste.path_t,
            "-f",
            ".ogg",
            ste.add_args,
        )

        if ste.combo == "album-->album":
            _assert_file_tree(ste.path_t, ste.path_e, depth=1)
            assert ste.path_t.joinpath("audio_at_artist_level.ogg").is_file()
            assert not ste.path_t.joinpath("album_good_(2018)").exists()
        elif ste.combo == "folder-->folder_part":
            _assert_file_tree(ste.path_t, ste.path_e)
            assert cmp_files_hash_and_time(org_file_list, saved_file_info) == 7
        else:
            _assert_file_tree(ste.path_t, ste.path_e)


class TestMmusicc:
    """Test complete program"""

    @pytest.mark.parametrize(
        "opt",
        [
            None,
            "--lazy",
            "--delete-existing-metadata",
            "--lazy --delete-existing-metadata",
        ],
    )
    def test_default(
        self, dir_lib_a_flac, dir_lib_c_ogg, dir_lib_test, dir_lib_b_ogg, opt
    ):
        """test the program for the default case it is made for with most used
            parameters
        """
        copy_tree(str(dir_lib_c_ogg), str(dir_lib_test))
        org_file_list = get_file_list_tree(dir_lib_test)
        saved_file_info = save_files_hash_and_mtime(dir_lib_test, touch=True)
        _assert_run_mmusicc(
            "--source", dir_lib_a_flac, "--target", dir_lib_test, "-f .ogg", opt
        )

        # check that missing files are created
        _assert_file_tree(dir_lib_test, dir_lib_b_ogg)
        path_changed = dir_lib_test.joinpath(
            "artist_quodlibet/album_bar_-_single_(2020)/01_track1.ogg"
        )
        metadata = Metadata(str(path_changed))
        assert metadata.get_tag("album") == "Bar - Single"
        assert metadata.get_tag("date") == "2020"
        assert metadata.get_tag("artist") == "Quod Libet"

        cmp_th = cmp_files_hash_and_time(org_file_list, saved_file_info)
        # see test_folder_folder_part for explanation
        if opt is None:
            assert cmp_th == 10107
        else:
            assert cmp_th == 30307

        if opt and "--delete-existing-metadata" in opt:
            assert len(metadata.unprocessed_tag) == 0
            if "--lazy" in opt:
                assert metadata.get_tag("composer") == "should not be here"
            else:
                assert metadata.get_tag("composer") is None
        else:
            assert len(metadata.unprocessed_tag) > 0
            assert metadata.get_tag("composer") == "should not be here"

        # run a second time to ensure its deterministic
        saved_file_info = save_files_hash_and_mtime(dir_lib_test, touch=True)
        _assert_run_mmusicc(
            "--source", dir_lib_a_flac, "--target", dir_lib_test, "-f .ogg", opt
        )
        # check no file was modified, first run should have done all
        assert cmp_files_hash_and_time(dir_lib_test, saved_file_info) == 11

    def test_custom_config_path(self, dir_lib_a_flac, dir_lib_test, dir_orig_data):
        """run mmusicc with testing config file, which is not the default one

            because init_allocationmap will be skipped when list_tags is
            already filled. It is cleared so the init will load the new
            config file. Since the map is force initialized module wise by
            the fixture it is reset to the default.
        """
        import mmusicc.util.allocationmap as am

        am.list_tags = list()
        abs_path = str(dir_orig_data.joinpath("metadata_config.yaml"))
        _assert_run_mmusicc(
            "--only-meta",
            "--source",
            dir_lib_a_flac,
            "--target",
            dir_lib_test,
            "--path-config",
            abs_path,
            "-f ogg",
        )
        assert len(am.list_tags) == 23
        am.list_tags = list()
        _assert_run_mmusicc(
            "--only-meta",
            "--source",
            dir_lib_a_flac,
            "--target",
            dir_lib_test,
            "-f ogg",
        )
        assert len(am.list_tags) == 16


def _assert_file_tree(tree_a, tree_b, depth=None) -> (list, list):
    """asserts if tree structure of sub-files and directories is identical

        and returns a lists of the compared files as tuple (files_a, files_b).
    """
    files_a, base_length_a = get_file_list_tree(tree_a, depth=depth, ret_base=True)
    files_b, base_length_b = get_file_list_tree(tree_b, depth=depth, ret_base=True)

    for i in range(len(files_a)):
        com_parts_a = files_a[i].parts[base_length_a:]
        com_parts_b = files_b[i].parts[base_length_b:]
        assert com_parts_a == com_parts_b
    return files_a, files_b


def _cmd_mmusicc(*args):
    """create and return a proper command list from a mixed input list

        strings will be splited at whitespaces, to pass a string like object
        without splitting (eg. for ffmpeg args) pass it as bytes.
    """
    final_cmd = list()
    if isinstance(args, pathlib.Path):
        args = str(args)
    for arg in args:
        if isinstance(arg, str):
            if arg == "":
                continue
            final_cmd.extend(arg.split())
        elif isinstance(arg, bytes):
            final_cmd.append(arg.decode())
        elif isinstance(arg, pathlib.Path):
            final_cmd.append(str(arg))
        elif arg is None:
            continue
        else:
            for a in arg:
                final_cmd.extend(_cmd_mmusicc(a))
    return final_cmd


def _assert_run_mmusicc(*args):
    """runs mmusicc with the given arguments. args will be preprocessed with
        cmd_command().
    """
    with pytest.raises(SystemExit) as excinfo:
        MmusicC(_cmd_mmusicc(*args))
    assert excinfo.value.code == 0


def test_assert_file_tree(dir_lib_b_ogg, dir_lib_c_ogg):
    files_a, files_b = _assert_file_tree(dir_lib_b_ogg, dir_lib_b_ogg)
    assert len(files_a) == 11
    assert len(files_b) == 11
    files_a, files_b = _assert_file_tree(dir_lib_b_ogg, dir_lib_b_ogg, 2)
    assert len(files_a) == 1
    assert len(files_b) == 1
    with pytest.raises(AssertionError):
        _assert_file_tree(dir_lib_c_ogg, dir_lib_b_ogg)


def test_cmd_mmusicc():
    cmd = _cmd_mmusicc("-fuu bar", ["--Hello", "World"], ["-qood libet"])
    assert cmd == ["-fuu", "bar", "--Hello", "World", "-qood", "libet"]
